<?php

namespace Drupal\h5p_challenge\Controller;

use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Database\Driver\mysql\Connection;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use Drupal\Component\Uuid\Php as Uuid;
use Drupal\h5p_challenge\H5PChallengeServiceInterface;
use Symfony\Component\HttpFoundation\Request;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Mail\MailManagerInterface;
use Drupal\Core\Url;
use Drupal\h5p_challenge\FetchClass\H5PChallenge;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Link;
use Drupal\language\ConfigurableLanguageManagerInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\h5p\Entity\H5PContent;

/**
 * Class H5PChallengeController.
 */
class H5PChallengeController extends ControllerBase {

  /**
   * Drupal\Core\Database\Driver\mysql\Connection definition.
   *
   * @var \Drupal\Core\Database\Driver\mysql\Connection
   */
  protected $database;

  /**
   * Uuid service
   *
   * @var \Drupal\Component\Uuid\Php
   */
  protected $uuid;

  /**
   * Date formatter
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * Mail manager
   *
   * @var \Drupal\Core\Mail\MailManagerInterface
   */
  protected $mailManager;

  /**
   * Challenge service
   *
   * @var \Drupal\h5p_challenge\H5PChallengeServiceInterface
   */
  protected $challengeService;

  /**
   * Language manager
   *
   * @var \Drupal\language\ConfigurableLanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Current user
   *
   * @var AccountProxyInterface $currentUser
   */
  protected $currentUser;

  /**
   * Constructs a new H5PChallengeController object.
   */
  public function __construct(Connection $database, Uuid $uuid, H5PChallengeServiceInterface $challengeService, DateFormatterInterface $dateFormatter, MailManagerInterface $mailManager, ConfigurableLanguageManagerInterface $languageManager, AccountProxyInterface $currentUser) {
    $this->database = $database;
    $this->uuid = $uuid;
    $this->challengeService = $challengeService;
    $this->dateFormatter = $dateFormatter;
    $this->mailManager = $mailManager;
    $this->languageManager = $languageManager;
    $this->currentUser = $currentUser;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('uuid'),
      $container->get('h5p.challenge.default'),
      $container->get('date.formatter'),
      $container->get('plugin.manager.mail'),
      $container->get('language_manager'),
      $container->get('current_user')
    );
  }

  /**
   * Start new challenge
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function createNew(Request $request) {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $title = filter_input(INPUT_POST, 'title');
    $email = filter_input(INPUT_POST, 'email', FILTER_VALIDATE_EMAIL);
    $duration = filter_input(INPUT_POST, 'duration', FILTER_VALIDATE_INT);
    $captcha = filter_input(INPUT_POST, 'g-recaptcha-response');
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $title && $email !== NULL && $email !== FALSE && $duration !== NULL && $duration !== FALSE && $captcha && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    if (! \Drupal::service('email.validator')->isValid($email)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Provided email address is invalid'),
      ]);
    }

    if ($duration < 1 && $duration > 168) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Provided duration is invalid'),
      ]);
    }

    if (!$this->challengeService->checkIfContentExists($content_id)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Content does not exist'),
      ]);
    }

    try {
      $captchaResponse = \Drupal::httpClient()->post(H5P_CHALLENGE_RECAPTCHA_VERIFY_ENDPOINT, [
        'form_params' => [
          'secret' => $this->challengeService->getRecaptchaSecretKey(),
          'response' => $captcha,
          'remoteip' => $request->getClientIp(),
        ],
      ]);
    } catch (\Exception $e) {
      \Drupal::logger('h5p_challenge')->error($e->getMessage());

      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Captcha service responded with an error'),
      ]);
    }

    if ($captchaResponse->getStatusCode() !== 200 || $captchaResponse->getReasonPhrase() !== 'OK') {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Captcha service responded with an error'),
      ]);
    }

    $parsedResponse = json_decode($captchaResponse->getBody(), TRUE);
    if ($parsedResponse['success'] !== TRUE) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong or unsuitable captcha'),
      ]);
    }

    $started = time();
    $finished = $started + ($duration * 60 * 60);
    $code = $this->challengeService->generateUniqueChallengeCode($content_id);
    $langcode = $this->languageManager->getCurrentLanguage()->getId();
    $uuid = $this->uuid->generate();
    $url = Url::fromRoute('h5p_challenge.h5p_challenge_controller_results', [
      'uuid' => $uuid,
    ], [
      'absolute' => TRUE,
    ])->toString();

    // TODO Handle failures?
    $this->database->insert('h5p_challenge')->fields([
      'uuid' => $uuid,
      'content_id' => $content_id,
      'title' => $title,
      'email' => $this->currentUser->isAuthenticated() ? $this->currentUser->getEmail() : $email,
      'started' => $started,
      'finished' => $finished,
      'code' => $code,
      'langcode' => $langcode,
      'user_id' => $this->currentUser->isAuthenticated() ? $this->currentUser->id() : NULL,
    ])->execute();

    if ($this->currentUser->isAuthenticated()) {
      $message = t("You have created a new challenge: @title.\nThe title of a resource for this challenge is: @ctitle.\n\nThe challenge will be running from @started to @finished (the duration would be @hours hours).\n\nChallenge code to participate is: @code\n\nYou can track results at: @results\nAnother notification containing results would be sent to you soon after the challenge is over.\nThe results page will never expire and you can always access it from the list of your challenges.\n\n\nBest regards,\nChallenge team.", [
        '@title' => $title,
        '@ctitle' => H5PChallenge::getContentTitleById($content_id),
        '@started' => $this->dateFormatter->format($started, 'long'),
        '@finished' => $this->dateFormatter->format($finished, 'long'),
        '@hours' => $duration,
        '@code' => $code,
        '@results' => $url,
      ]);
    } else {
      $message = t("You have created a new challenge: @title.\nThe title of a resource for this challenge is: @ctitle.\n\nThe challenge will be running from @started to @finished (the duration would be @hours hours).\n\nChallenge code to participate is: @code\n\nYou can track results at: @results\nAnother notification containing results would be sent to you soon after the challenge is over.\nThe results link would still be active for at least 36 hours after the challenge ends.\n\n\nBest regards,\nChallenge team.", [
        '@title' => $title,
        '@ctitle' => H5PChallenge::getContentTitleById($content_id),
        '@started' => $this->dateFormatter->format($started, 'long'),
        '@finished' => $this->dateFormatter->format($finished, 'long'),
        '@hours' => $duration,
        '@code' => $code,
        '@results' => $url,
      ]);
    }
    $params = [
      'body' => (string) $message,
    ];
    $this->mailManager->mail('h5p_challenge', 'new_challenge_created', $email, $langcode, $params, NULL, TRUE);

    return new JsonResponse([
      'success' => TRUE,
      'code' => $code,
      'url' => $url,
    ]);
  }

  /**
   * Join existing challenge
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function startPlaying() {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $name = filter_input(INPUT_POST, 'name');
    $code = filter_input(INPUT_POST, 'code');
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $name && $code && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    $challenge = $this->challengeService->getChallengeByContentIdAndCode($content_id, $code);

    if ($challenge && $challenge->isActive()) {
      $uuid = $this->uuid->generate();

      $this->database->insert('h5p_challenge_points')->fields([
        'challenge_uuid' => $challenge->uuid,
        'player_uuid' => $uuid,
        'name' => $this->currentUser->isAuthenticated() ? $this->currentUser->getDisplayName() : $name,
        'started' => time(),
        'user_id' => $this->currentUser->isAuthenticated() ? $this->currentUser->id() : NULL,
      ])->execute();

      $data = [
        'uuid' => $uuid,
        'challenge' => [
          'title' => $challenge->title,
          'started' => $challenge->started,
          'finished' => $challenge->finished,
        ],
      ];

      user_cookie_save(['h5p-challenge-for-' . $content_id => json_encode($data),]);

      return new JsonResponse(array_merge($data, [
        'success' => TRUE,
      ]));
    } else if ($challenge) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Challenge is no longer active'),
      ]);
    } else {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('No challenge could be found for provided code'),
      ]);
    }

    return new JsonResponse([
      'success' => FALSE,
    ]);
  }

  /**
   * Store results for challenge participant
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response with possible additional data or message
   */
  public function setFinished() {
    $content_id = filter_input(INPUT_POST, 'contentId', FILTER_VALIDATE_INT);
    $uuid = filter_input(INPUT_POST, 'uuid');
    $score = filter_input(INPUT_POST, 'score', FILTER_VALIDATE_INT);
    $max_score = filter_input(INPUT_POST, 'maxScore', FILTER_VALIDATE_INT);
    $token = filter_input(INPUT_GET, 'token');

    if (! ($content_id && $uuid && $score !== NULL && $score !== FALSE && $max_score !== NULL && $max_score !== FALSE && $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Wrong usage'),
      ]);
    }

    /*if (! \H5PCore::validToken('result', $token)) {
      return new JsonResponse([
        'success' => FALSE,
        'message' => $this->t('Invalid security token'),
      ]);
    }*/

    $points = $this->challengeService->getChallengePointsByUuid($uuid);

    if ($points) {
      $challenge = $this->challengeService->getChallengeByUuid($points->challenge_uuid);

      if ($challenge && $challenge->isActive()) {
        $finished_time = time();

        $this->database->update('h5p_challenge_points')
        ->fields([
          'finished' => $finished_time,
          'points' => $score,
          'max_points' => $max_score,
        ])
        ->expression('tries_count', 'tries_count + :amount', [':amount' => 1,])
        ->condition('player_uuid', $uuid)
        ->execute();

        // Update cookie contents to indicate that score has been stored
        $cname = 'Drupal_visitor_h5p-challenge-for-' . $content_id;
        if (isset($_COOKIE[$cname])) {
          $data = json_decode($_COOKIE[$cname], TRUE);
          $data['finished'] = [
            'finished' => $finished_time,
            'score' => $score,
            'maxScore' => $max_score,
          ];
          user_cookie_save(['h5p-challenge-for-' . $content_id => json_encode($data),]);
        }

        return new JsonResponse([
          'success' => TRUE,
        ]);
      } else if ($challenge && !$challenge->isActive()) {
        return new JsonResponse([
          'success' => FALSE,
          'message' => $this->t('Challenge is no longer active'),
        ]);
      }
    }

    return new JsonResponse([
      'success' => FALSE,
    ]);
  }

  /**
   * Generates mastery level markup.
   *
   * @param  int    $points
   *   Points received
   * @param  int    $max_points
   *   Maximum possible points
   * @param  int    $tries_count
   *   Number of tries
   *
   * @return Markup
   *   Markup with colored round
   */
  private function masteryLevelMarkup(int $points, int $max_points, int $tries_count) : Markup {
    if ($max_points === 0 || $tries_count === 0) {
      return Markup::create('<span data-tippy-content="' . $this->t('Not finished yet.') . '" class="mastery-level"></span>');
    }

    $class = 'mastery-level';
    $score_percent = $points / $max_points * 100;

    if ($score_percent === 100 && $tries_count === 1) {
      $class .= ' green';
      $tippy = $this->t('Maximum score on first attempt.');
    } else if ($score_percent > 51 && $tries_count <= 2 ) {
      $class .= ' orange';
      $tippy = $this->t('Score above 51% with two or less attempts.');
    } else if ($score_percent < 50 || $tries_count > 2){
      $class .= ' red';
      $tippy = $this->t('Score below 50% or more than two attempts.');
    }

    return Markup::create('<span data-tippy-content="' . $tippy . '" class="' . $class . '"></span>');
  }

  /**
   * Converts time taken to minutes.
   *
   * @param  int    $time_taken
   *   Time taken in seconds
   * @return string
   *   Time taken or empty string
   */
  private function timeTakenMarkup(int $time_taken) : string {
    if ($time_taken === 0) {
      return '';
    }

    $hours = 0;
    $minutes = 0;

    if ($time_taken > 3600) {
      $hours = floor($time_taken / 3600);
      $time_taken = $time_taken - $hours * 3600;
    }

    $minutes = ceil($time_taken / 60);

    if ($hours && $minutes) {
      return $this->t(':hoursh :minutesm', [
        ':hours' => $hours,
        ':minutes' => $minutes,
      ]);
    }

    return $this->t(':minutesm', [
      ':minutes' => $minutes,
    ]);
  }

  /**
   * Determines sub content with possibility of having a score. Only top level
   * children are considered, not questions within questions. Returns an array
   * with all embedded content that is considered to be questions.
   *
   * @param  H5PContent $entity
   *   Entity
   *
   * @return array
   *  An array with id = subContentId, type = library with version removed and
   *  title = metadata.title
   */
  private function getChallengeQuestions(H5PContent $entity) : array {
    $response = [];
    // These types were taken from H5P.Column codebase and are considered to be
    // tasks
    $types_with_score = [
      'H5P.ImageHotspotQuestion',
      'H5P.Blanks',
      'H5P.Essay',
      'H5P.SingleChoiceSet',
      'H5P.MultiChoice',
      'H5P.TrueFalse',
      'H5P.DragQuestion',
      'H5P.Summary',
      'H5P.DragText',
      'H5P.MarkTheWords',
      'H5P.MemoryGame',
      'H5P.QuestionSet',
      'H5P.InteractiveVideo',
      'H5P.CoursePresentation',
      'H5P.DocumentationTool',
    ];
    $json = $entity->getFilteredParameters();
    $iterator = new \RecursiveIteratorIterator(new \RecursiveArrayIterator(json_decode($json), \RecursiveIteratorIterator::SELF_FIRST));

    foreach($iterator as $key => $value) {
      if ($key === 'subContentId' && $iterator->getDepth() <= 3) {
        $si_data = $iterator->getInnerIterator()->getArrayCopy();
        $si_library = explode(' ', trim($si_data['library']))[0];

        if (in_array($si_library, $types_with_score)) {
          $response[$si_data['subContentId']] = [
            'id' => $si_data['subContentId'],
            'type' => $si_library,
            'title' => $si_data['metadata']->title,
          ];
        }
      }
    }

    return $response;
  }

  /**
   * Show challenge results page
   *
   * @return array
   *   Page structure
   */
  public function results(Request $request) {
    $challenge = $this->challengeService->getChallengeByUuid($request->attributes->get('uuid'));

    if ($challenge) {
      $response = [];
      $response['#title'] = $this->t('Challenge results: @title', [
        '@title' => $challenge->title,
      ]);

      $header = [
        $this->t('No.'),
        $this->t('Name'),
        $this->t('Mastery level'),
        $this->t('Total score'),
        $this->t('Time taken'),
        $this->t('Attempts'),
      ];

      /*$entity = H5PContent::load(35);

      if ($entity) {
        $questions = $this->getChallengeQuestions($entity);
        $no = 1;
        if (count($questions) > 0) {
          foreach ($questions as $question) {
            $header[] = Markup::create('<span data-tippy-content="' . $question['title'] . '">Q' . $no . '</span>');
            $no++;
          }
        }
      }*/

      $rows = [];

      $result = $this->database->select('h5p_challenge_points', 'cp')
      ->fields('cp')
      ->condition('cp.challenge_uuid', $challenge->uuid)
      ->orderBy('cp.started', 'ASC')
      ->execute();

      if ($result) {
        $no = 1;
        while($single = $result->fetchObject()) {
          $rows[] = [
            'data' => [
              'no' => $no . '.',
              'name' => $single->name,
              'mastery_level' => [
                'data' => $this->masteryLevelMarkup((int)$single->points, (int)$single->max_points, (int)$single->tries_count),
                'class' => 'mastery-level'
              ],
              'total_score' => ($single->points ?? 0) . '/' . ($single->max_points ?? 0),
              'time_taken' => $this->timeTakenMarkup($single->finished ? ($single->finished - $single->started) : 0), // XXX This should use the time reported by H5P xAPI
              'attempts' => $single->tries_count,
            ],
          ];
          $no++;
        }
      }

      $response['results-csv'] = [
        '#type' => 'link',
        '#title' => $this->t('Download results as CSV'),
        '#url' => Url::fromRoute('h5p_challenge.h5p_challenge_controller_results_csv', [
          'uuid' => $challenge->uuid,
        ], [
          'absolute' => TRUE,
        ]),
        '#attributes' => [
          'class' => ['btn', 'btn-primary', 'btn-xs',],
        ],
      ];

      $response['results-table'] = [
        '#theme' => 'table',
        '#rows' => $rows,
        '#header' => $header,
        '#attached' => [
          'library' => [
            'h5p_challenge/results',
          ],
        ],
      ];

      return $response;
    }

    throw new NotFoundHttpException();
  }

  /**
   * Serve challenge results to be downloaded as CSV
   *
   * @return \Symfony\Component\HttpFoundation\Result
   *   Challenge results file
   */
  public function resultsCsv(Request $request) {
    $challenge = $this->challengeService->getChallengeByUuid($request->attributes->get('uuid'));

    if ($challenge) {
      $response = new Response($this->challengeService->getChallengeReultsCsv($challenge));

      $response->headers->set('Content-Type', 'text/csv');
      $response->headers->set('Content-Disposition','attachment; filename="results.csv"');

      return $response;
    }

    throw new NotFoundHttpException();
  }

  /**
   * List of currrently present challenges
   *
   * @return array
   *   Page structure
   */
  public function reportsList() {
    $query = $this->database->select('h5p_challenge', 'c', [
      'fetch' => H5PChallenge::class,
    ]);
    $query->fields('c', ['uuid', 'content_id', 'title', 'email', 'started', 'finished', 'code', 'results_sent', 'langcode',]);
    $pager = $query->extend('Drupal\Core\Database\Query\PagerSelectExtender')->limit(25);
    $result = $pager->execute();

    $header = [
      $this->t('Title'),
      $this->t('Started'),
      $this->t('Finished'),
      $this->t('Code'),
      $this->t('Results sent'),
      $this->t('Language'),
    ];
    $rows = [];

    foreach ($result->fetchAll() as $challenge) {
      $rows[] = [
        'data' => [
          'title' => Link::createFromRoute(
            $challenge->title,
            'h5p_challenge.h5p_challenge_controller_results',
            ['uuid' => $challenge->uuid],
            [
              'attributes' => [
                'target' => '_blank',
              ]
            ]
          ),
          'started' => $this->dateFormatter->format($challenge->started, 'long'),
          'finished' => $this->dateFormatter->format($challenge->finished, 'long'),
          'code' => $challenge->code,
          'results_sent' => ($challenge->results_sent) ? $this->t('Yes') : $this->t('No'),
          'language' => $this->languageManager->getLanguageName($challenge->langcode),
        ],
      ];
    }

    $response['challenges'] = [
      '#type' => 'table',
      '#caption' => $this->t('Challenges'),
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => $this->t('No challenges found'),
    ];

    $response['pager'] = [
      '#type' => 'pager',
    ];

    return $response;
  }

  /**
   * List of challenges created by user
   *
   * @return array
   *   Page structure
   */
  public function mine() {
    $query = $this->database->select('h5p_challenge', 'c', [
      'fetch' => H5PChallenge::class,
    ]);
    $query->condition('c.user_id', $this->currentUser->id(), '=');
    $query->fields('c', ['uuid', 'content_id', 'title', 'email', 'started', 'finished', 'code', 'results_sent', 'langcode',]);
    $pager = $query->extend('Drupal\Core\Database\Query\PagerSelectExtender')->limit(25);
    $result = $pager->execute();

    $header = [
      $this->t('Title'),
      $this->t('Started'),
      $this->t('Finished'),
      $this->t('Code'),
      $this->t('Results sent'),
      $this->t('Language'),
    ];
    $rows = [];

    foreach ($result->fetchAll() as $challenge) {
      $rows[] = [
        'data' => [
          'title' => Link::createFromRoute(
            $challenge->title,
            'h5p_challenge.h5p_challenge_controller_results',
            ['uuid' => $challenge->uuid],
            []
          ),
          'started' => $this->dateFormatter->format($challenge->started, 'long'),
          'finished' => $this->dateFormatter->format($challenge->finished, 'long'),
          'code' => $challenge->code,
          'results_sent' => ($challenge->results_sent) ? $this->t('Yes') : $this->t('No'),
          'language' => $this->languageManager->getLanguageName($challenge->langcode),
        ],
      ];
    }

    $response['challenges'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $rows,
      '#empty' => $this->t('No challenges found'),
    ];

    $response['pager'] = [
      '#type' => 'pager',
    ];

    return $response;
  }

}
